# -*- coding: utf-8 -*-
"""
Created on 2018-07-01, Last Modified on 2020-02-10;
@author Yi.Q
"""
"""
Metrics to assess the performance on a classification model
Measures are from: https://en.wikipedia.org/wiki/Confusion_matrix;

———————————————————————————————————————————————————————————————————————————
         |                                       actual condition          |
         |—————————————————————————————————————————————————————————————————
         |                    |    actual positive    |  actual negative   |
         |—————————————————————————————————————————————————————————————————
Predicted| predicted positive |     true positive     |   false positive   |
condition|—————————————————————————————————————————————————————————————————
         | predicted negative |     false negative    |   true negative    |
—————————|—————————————————————————————————————————————————————————————————

Terms and calculations:
    thresholds: pre-set thresholds on which the follower metrics are calculated;
    TP: true positive, power, the number of obs. that is conditioned positive and predicted positive;
    FN: false negative, type II error, the number of obs. that is conditioned positive and predicted negative;
    FP: false positive, type I error, the number of obs. that is conditioned negative and predicted positive;
    TN: true negative, the number of obs. that is conditioned negative and predicted negative;

    Recall: TPR, probability of detecting true positive condition, = TP / (TP + FN);
    FNR: miss rate, = FN / (TP + FN);
    FPR: probability of false alarm, = FP / (FP + TN);
    TNR: tue negative rate, = TN / (FP + TN);

    Precision: Positive predictive value(PPV), = TP / (TP + FP);
    FDR: false discovery rate, FP / (TP + FP);
    FOR: false omission rate, FN / (FN + TN);
    NPV: negative predictive value, TN / (FN + TN);

    prevalence: = condition positive / total population;
    Accuracy: = (TP + TN) / (TP + FN + FP + TN);
    LR+: = TPR / FPR;
    LR-: = FNR / TNR;
    DOR: diagnostic odds ratio, = LR+ / LR-;
    F1: 2 / ((1/Recall) + (1/Precision));
"""
from typing import Union, Tuple, Optional

import numpy as np

from kolibri.classifier_evaluator.styles.format_confusion_matrix_by_prob_df import \
    convert_confusion_matrix_by_prob_to_table_with_reformat_precision

_FULL_METRICS = ['TP', 'FN', 'FP', 'TN',
                 'Recall', 'FNR', 'FPR', 'TNR', 'Precision', 'FOR', 'FDR', 'NPV',
                 'Prevalence', 'Accuracy', 'LR+', 'LR-', 'DOR', 'F1']

_FULL_METRICS_DUAL = ['Open_Count', 'Open_Positive', 'Open_Negative', 'Open_Prevalence']

_DEFAULT_POS_LABEL = True


def accuracy_rate(true: np.ndarray, predicted: np.ndarray) -> float:
    """
    calculate accuracy rate (ACC):
        ( true positive + true negative ) / sample size

    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted: numpy.ndarray(shape=(m), ), an array of predicted classes;
    :return: float, fraction, accuracy rate
    """
    sample_size = len(true)
    return (predicted == true).sum() / sample_size


def accuracy_rate_by_prob(true: np.ndarray, predicted_prob: np.ndarray,
                          threshold: float, pos_label: Union[str, bool, int] = _DEFAULT_POS_LABEL) -> float:
    """
    accuracy rate for binary classification according to a given threshold;

    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted_prob: numpy.ndarray(shape=(m), ),
        an array of predicted probabilities of being the positive class;
    :param threshold: float, [0, 1], the threshold set on predicted probabilities
        such that any predicted probability greater or equal to the threshold will be classified as the positive class,
        e.g., predicted_class(i | predicted_prob_i >= threshold) = positive;
    :param pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
    :return: float, fraction, accuracy rate
    """
    # convert true series to positive - negative class
    true = (true == pos_label)
    predicted = (predicted_prob >= threshold)

    return accuracy_rate(true=true, predicted=predicted)


def _condition_counter(true: np.ndarray,
                       predicted: np.ndarray,
                       true_condition: Union[bool, int, str, list, tuple, set],
                       predicted_condition: Union[bool, int, str, list, tuple, set]) -> int:
    """
    count the number of instances such that their true value being true_condition and predicted being predicted_condition;
    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted: numpy.ndarray(shape=(m), ), an array of predicted classes;
    :param true_condition: [bool, int, str, list, tuple, set], the true condition(s) that is being count;
    :param predicted_condition: [bool, int, str, list, tuple, set], the predicted condition(s) that is being count;
    :return: total count of the instance with the joined condition, int;
    """
    if isinstance(true_condition, (str, bool, int)):
        true_condition = [true_condition]
    elif isinstance(true_condition, set):
        true_condition = list(true_condition)
    elif isinstance(true_condition, (list, tuple)):
        pass
    else:
        raise TypeError(f"Type of true conditions {type(true_condition)} is not acceptable.")

    if isinstance(predicted_condition, (str, bool, int)):
        predicted_condition = [predicted_condition]
    elif isinstance(predicted_condition, set):
        predicted_condition = list(predicted_condition)
    elif isinstance(predicted_condition, (list, tuple)):
        pass
    else:
        raise TypeError(f"Type of true conditions {type(predicted_condition)} is not acceptable.")

    return (np.isin(true, true_condition) & np.isin(predicted, predicted_condition)).sum().item()


def prevalence(series: np.ndarray,
               condition: Union[bool, int, str, list, tuple, set],
               rate: bool = False) -> Union[int, float]:
    """
    count the number of instances with (one of) the given condition(s);
    :param series: numpy.ndarray(shape=(m), ) containing different conditions;
    :param condition: [bool, int, str, list, tuple, set], a selected condition;
    :param rate: bool, return prevalence rate instead of count;
    :return: the count of the instances with the condition(s);
    """
    if isinstance(condition, (str, bool, int)):
        condition = [condition]
    elif isinstance(condition, set):
        condition = list(condition)
    elif isinstance(condition, (list, tuple)):
        pass
    else:
        raise TypeError(f"Type of true conditions {type(condition)} is not acceptable.")

    if rate:
        return np.isin(series, condition).sum() / series.size if series.size > 0 else 0
    else:
        return np.isin(series, condition).sum()


def recall_rate(true: np.ndarray,
                predicted: np.ndarray,
                condition: Union[bool, int, str, list, tuple, set]) -> float:
    """
    compute recall rate, refer module description for calculation;
    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted: numpy.ndarray(shape=(m), ), an array of predicted classes;
    :param condition: [bool, int, str, list, tuple, set], the true condition(s) that is being count;
    :return: recall rate, float;
    """
    joint_condition_count = _condition_counter(true=true,
                                               predicted=predicted,
                                               true_condition=condition,
                                               predicted_condition=condition)

    prevalence_count = prevalence(series=true,
                                  condition=condition,
                                  rate=False)

    recall = joint_condition_count / prevalence_count if prevalence_count > 0 else 0

    return recall


def recall_rate_by_prob(true: np.ndarray,
                        predicted_prob: np.ndarray,
                        threshold: float = 0.5,
                        pos_label: Union[bool, int, str] = _DEFAULT_POS_LABEL) -> float:
    """
    compute recall rate when predicted prob is given, refer module description for calculation;
    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted_prob: numpy.ndarray(shape=(m), ), an array of predicted probability of being positive classe;
    :param threshold: float, the threshold over which predicted probabilities will be counted as positive;
    :param pos_label: [bool, int, str], the true condition(s) that is being count;
    :return: recall rate, float;
    """

    true = true == pos_label
    predicted = predicted_prob >= threshold

    recall = recall_rate(true=true,
                         predicted=predicted,
                         condition=True)

    return recall


def precision_rate(true: np.ndarray,
                   predicted: np.ndarray,
                   condition: Union[bool, int, str, list, tuple, set]) -> float:
    """
    compute precision rate, refer module description for calculation;
    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted: numpy.ndarray(shape=(m), ), an array of predicted classes;
    :param condition: [bool, int, str, list, tuple, set], the true condition(s) that is being count;
    :return: precision rate, float;
    """
    joint_condition_count = _condition_counter(true=true,
                                               predicted=predicted,
                                               true_condition=condition,
                                               predicted_condition=condition)

    prevalence_count = prevalence(series=predicted,
                                  condition=condition,
                                  rate=False)

    precision = joint_condition_count / prevalence_count if prevalence_count > 0 else 0

    return precision


def precision_rate_by_prob(true: np.ndarray,
                           predicted_prob: np.ndarray,
                           threshold: float = 0.5,
                           pos_label: Union[bool, int, str] = _DEFAULT_POS_LABEL) -> float:
    """
    compute precision rate when predicted prob is given, refer module description for calculation;
    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted_prob: numpy.ndarray(shape=(m), ), an array of predicted probability of being positive classe;
    :param threshold: float, the threshold over which predicted probabilities will be counted as positive;
    :param pos_label: [bool, int, str], the true condition(s) that is being count;
    :return: precision rate, float;
    """

    true = true == pos_label
    predicted = predicted_prob >= threshold

    precision = precision_rate(true=true,
                               predicted=predicted,
                               condition=True)

    return precision


def confusion_matrix(true: np.ndarray, predicted: np.ndarray, normalize: Union[bool, str] = False) -> Union[
    tuple, dict]:
    """
    compute confusion matrix

        terms: pair condition: joined condition of both observed condition and predicted condition, e.g., (True, False);
               true condition: observed condition of an instance, e.g., True;
               predicted condition: predicted condition of an instance, e.g., True;
    :param true: true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted: numpy.ndarray(shape=(m), ), an array of predicted classes;
    :param normalize: [bool, 'true', 'predicted'], if the confusion matrix needs to be normalised:
        True: default, normalized along true classes; False: not normalized;
        'true': normalized along true classes; 'predicted': normalized along predicted classes;
        e.g., normalisation along true classes: normalized_condition (for i) = count(predicted = i, true = i) / (true = i)
    :return: dict, a confusion matrix:
        key: tuple, (true_class, predicted_class);
        value: [int, float], the number of instances of that pair condition (true_class, predicted_class), or
            normalisation of the number of the instances of that pair condition;
    """
    labels = [label for label in set(true).union(set(predicted))]  # .item convert numpy dtypes to native types;

    confusion_matrix_dict = dict()
    for true_label in labels:
        for predicted_label in labels:
            confusion_matrix_dict[(true_label, predicted_label)] = \
                _condition_counter(true=true, predicted=predicted,
                                   true_condition=true_label, predicted_condition=predicted_label)

    if normalize:
        normalize_factor = 0 if normalize != 'predicted' else 1
        normalized_confusion_matrix_dict = normalize_confusion_matrix(confusion_matrix_dict=confusion_matrix_dict,
                                                                      normalize_index=normalize_factor)
        return confusion_matrix_dict, normalized_confusion_matrix_dict
    else:
        return confusion_matrix_dict


def normalize_confusion_matrix(confusion_matrix_dict: dict, normalize_index: int = 0) -> dict:
    """
    normalize a confusion matrix;

    :param confusion_matrix_dict: dict, a confusion matrix, e.g.,
        {('a', 'a'): 1, ('a', 'b'): 2, ('b', 'a'): 3, ('b', 'b'): 4};
    :param normalize_index: int, [0, 1], the position index of the confusion matrix dict key (tuple) where normalization
    is performed along;
    :return: a normalized confusion matrix;
    """
    labels = set([key[0] for key in confusion_matrix_dict.keys()])
    normalized_confusion_matrix_dict = dict()

    sum_dict = {label: 0 for label in labels}
    for condition, condition_count in confusion_matrix_dict.items():
        sum_dict[condition[normalize_index]] += condition_count

    for condition, condition_count in confusion_matrix_dict.items():
        normalized_confusion_matrix_dict[condition] = \
            condition_count / sum_dict[condition[normalize_index]] if sum_dict[condition[normalize_index]] else 0

    return normalized_confusion_matrix_dict


def confusion_matrix_by_prob(true: np.ndarray,
                             predicted_prob: np.ndarray,
                             thresholds: Optional[Union[list, tuple, np.ndarray]] = None,
                             pos_label: Union[bool, str, int] = _DEFAULT_POS_LABEL,
                             output_metrics: Optional[list] = None,
                             table: bool = True,
                             **kwargs):
    """
    confusion matrix for binary classification according to a given set of thresholds;

    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted_prob: numpy.ndarray(shape=(m), ),
        an array of predicted probabilities of being the positive class;
    :param thresholds: [list, tuple, np.array, None] the thresholds set on predicted probabilities
        such that any predicted probability greater or equal to the threshold will be classified as the positive class;
    :param pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
    :param output_metrics: [list, None], metrics to be outputted if selected;
    :param table: bool, if exported as a pd table table;
    :param kwargs:
        :param metric_order: [list, None], if table is selected to be the output, metric order specifies the order of
            metrics presented in the table;
    :return: dict, a set of confusion matrices, {threshold: {metric_name: metric_value, ...}, ...};
    """
    # convert true series to positive series
    true = true == pos_label

    # select output:
    if isinstance(output_metrics, list):
        for selected_metric in output_metrics:
            if selected_metric not in _FULL_METRICS:
                raise KeyError(f"metric {selected_metric} is not recognized.")
    elif output_metrics == 'confusion':
        output_metrics = ['TP', 'FN', 'FP', 'TN',
                          'Recall', 'FNR', 'FPR', 'TNR', 'Precision', 'FOR', 'FDR', 'NPV',
                          'Prevalence', 'Accuracy']
    else:
        output_metrics = _FULL_METRICS

    metrics_by_thresholds = dict()
    for threshold in thresholds:
        metrics_by_threshold = dict()
        predicted = predicted_prob >= threshold
        confusion_matrix_dict = confusion_matrix(true=true, predicted=predicted, normalize=False)

        confusion_matrix_nor_true = normalize_confusion_matrix(confusion_matrix_dict=confusion_matrix_dict,
                                                               normalize_index=0)

        confusion_matrix_nor_predicted = normalize_confusion_matrix(confusion_matrix_dict=confusion_matrix_dict,
                                                                    normalize_index=1)

        if 'TP' in output_metrics:
            metrics_by_threshold['TP'] = confusion_matrix_dict[(True, True)]

        if 'FN' in output_metrics:
            metrics_by_threshold['FN'] = confusion_matrix_dict[(True, False)]

        if 'FP' in output_metrics:
            metrics_by_threshold['FP'] = confusion_matrix_dict[(False, True)]

        if 'TN' in output_metrics:
            metrics_by_threshold['TN'] = confusion_matrix_dict[(False, False)]

        if 'Recall' in output_metrics:
            metrics_by_threshold['Recall'] = confusion_matrix_nor_true[(True, True)]

        if 'FNR' in output_metrics:
            metrics_by_threshold['FNR'] = confusion_matrix_nor_true[(True, False)]

        if 'FPR' in output_metrics:
            metrics_by_threshold['FPR'] = confusion_matrix_nor_true[(False, True)]

        if 'TNR' in output_metrics:
            metrics_by_threshold['TNR'] = confusion_matrix_nor_true[(False, False)]

        if 'Precision' in output_metrics:
            metrics_by_threshold['Precision'] = confusion_matrix_nor_predicted[(True, True)]

        if 'FOR' in output_metrics:
            metrics_by_threshold['FOR'] = confusion_matrix_nor_predicted[(True, False)]

        if 'FDR' in output_metrics:
            metrics_by_threshold['FDR'] = confusion_matrix_nor_predicted[(False, True)]

        if 'NPV' in output_metrics:
            metrics_by_threshold['NPV'] = confusion_matrix_nor_predicted[(False, False)]

        if 'Prevalence' in output_metrics:
            metrics_by_threshold['Prevalence'] = \
                (confusion_matrix_dict[(True, True)] + confusion_matrix_dict[(True, False)]) / sum(
                    confusion_matrix_dict.values())

        if 'Accuracy' in output_metrics:
            metrics_by_threshold['Accuracy'] = \
                (confusion_matrix_dict[(True, True)] + confusion_matrix_dict[(False, False)]) / sum(
                    confusion_matrix_dict.values())

        if 'LR+' in output_metrics:
            # positive likelihood ratio:
            try:
                metrics_by_threshold['LR+'] = confusion_matrix_nor_true[(True, True)] / confusion_matrix_nor_true[
                    (False, True)]
            except ZeroDivisionError:
                metrics_by_threshold['LR+'] = '-'

        if 'LR-' in output_metrics:
            # negative likelihood ratio:
            try:
                metrics_by_threshold['LR-'] = confusion_matrix_nor_true[(True, False)] / confusion_matrix_nor_true[
                    (False, False)]
            except ZeroDivisionError:
                metrics_by_threshold['LR-'] = '-'

        if 'DOR' in output_metrics:
            # diagnostic odds ratio:
            try:
                metrics_by_threshold['DOR'] = (confusion_matrix_nor_true[(True, True)] / confusion_matrix_nor_true[
                    (False, True)]) / \
                                              (confusion_matrix_nor_true[(True, False)] / confusion_matrix_nor_true[
                                                  (False, False)])
            except ZeroDivisionError:
                metrics_by_threshold['DOR'] = '-'

        if 'F1' in output_metrics:
            # F1 score:
            try:
                metrics_by_threshold['F1'] = 2 * (
                        confusion_matrix_nor_true[(True, True)] * confusion_matrix_nor_predicted[(True, True)]) / \
                                             (confusion_matrix_nor_true[(True, True)] + confusion_matrix_nor_predicted[
                                                 (True, True)])
            except ZeroDivisionError:
                metrics_by_threshold['F1'] = '-'

        metrics_by_thresholds[threshold] = metrics_by_threshold

    if table:
        if 'metric_order' in kwargs:
            metric_order = kwargs['metric_order']
        else:
            metric_order = None
        metrics_by_thresholds = \
            convert_confusion_matrix_by_prob_to_table_with_reformat_precision(
                metrics_by_thresholds=metrics_by_thresholds,
                metric_order=metric_order)

    return metrics_by_thresholds


def confusion_matrix_by_prob_with_dual_thresholds(true: np.ndarray,
                                                  predicted_prob: np.ndarray,
                                                  threshold_bounds: Union[list, tuple, set],
                                                  pos_label: Union[bool, str, int] = _DEFAULT_POS_LABEL,
                                                  output_metrics: Optional[list] = None,
                                                  unclassified_output_metrics: Optional[list] = None,
                                                  table: bool = True,
                                                  **kwargs):
    # modify threshold bound in case of only one threshold bound given:
    if isinstance(threshold_bounds, tuple) and isinstance(threshold_bounds[0], float):
        threshold_bounds = [threshold_bounds]

    # select output:
    if isinstance(unclassified_output_metrics, list):
        for selected_metric in unclassified_output_metrics:
            if selected_metric not in _FULL_METRICS_DUAL:
                raise KeyError(f"metric {selected_metric} is not recognized.")
    else:
        unclassified_output_metrics = _FULL_METRICS_DUAL

    metrics_by_thresholds_dual = dict()
    for threshold_bound in threshold_bounds:
        # filtering texts below lower threshold and upper threshold:
        lower_threshold, upper_threshold = threshold_bound
        true_sub = true[(predicted_prob <= lower_threshold) | (predicted_prob >= upper_threshold)]
        predicted_sub = predicted_prob[(predicted_prob <= lower_threshold) | (predicted_prob >= upper_threshold)]
        predicted_sub = (predicted_sub >= upper_threshold)  # all predicted_sub over upper threshold is set to 1:
        true_unclassified = true[(predicted_prob > lower_threshold) & (predicted_prob < upper_threshold)]

        metrics_by_thresholds = confusion_matrix_by_prob(true=true_sub,
                                                         predicted_prob=predicted_sub,
                                                         thresholds=[0.5],
                                                         pos_label=pos_label,
                                                         output_metrics=output_metrics,
                                                         table=False)

        # unpack structure:
        metrics_by_threshold = metrics_by_thresholds[0.5]  # default set to 0.5;

        # additional metrics:
        if 'Open_Count' in unclassified_output_metrics:
            metrics_by_threshold['Open_Count'] = len(true_unclassified)

        if 'Open_Positive' in unclassified_output_metrics:
            metrics_by_threshold['Open_Positive'] = (true_unclassified == pos_label).sum().item()

        if 'Open_Negative' in unclassified_output_metrics:
            metrics_by_threshold['Open_Negative'] = (true_unclassified != pos_label).sum().item()

        if 'Open_Prevalence' in unclassified_output_metrics:
            try:
                metrics_by_threshold['Open_Prevalence'] = (true_unclassified == pos_label).sum().item() / len(
                    true_unclassified)
            except ZeroDivisionError:
                metrics_by_threshold['Open_Prevalence'] = '-'

        metrics_by_thresholds_dual[threshold_bound] = metrics_by_threshold

    if table:
        if 'metric_order' in kwargs:
            metric_order = kwargs['metric_order']
        else:
            metric_order = None
        metrics_by_thresholds_dual = \
            convert_confusion_matrix_by_prob_to_table_with_reformat_precision(
                metrics_by_thresholds=metrics_by_thresholds_dual,
                metric_order=metric_order)

    return metrics_by_thresholds_dual


def roc(true: np.ndarray, predicted_prob: np.ndarray, pos_label: Union[bool, str, int] = True) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray]:
    """
    Receiver operating characteristic (ROC):

    :param true: numpy.ndarray(shape=(m), ), an array of true classes;
    :param predicted_prob: numpy.ndarray(shape=(m), ),
        an array of predicted probabilities of being the positive class;
    :param pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
    :return: tuple, (thresholds, fpr, tpr), thresholds: numpy.ndarray, fpr: numpy.ndarray, tpr: numpy.ndarray;
    """
    true = true == pos_label

    # sort by predicted prob:
    sorted_index = np.argsort(predicted_prob)[::-1]
    true = true[sorted_index]
    predicted_prob = predicted_prob[sorted_index]

    tp = 0
    fp = 0
    condition_positive = true.sum()
    condition_negative = (1 - true).sum()

    # set thresholds at predicted_prob:
    thresholds = predicted_prob
    tpr, fpr = [], []
    for i in range(len(thresholds)):
        if true[i] == 1:
            tp += 1
        elif true[i] == 0:
            fp += 1

        tpr_i = tp / condition_positive  # consistent with confusion matrix by prob;
        fpr_i = fp / condition_negative

        tpr.append(tpr_i)
        fpr.append(fpr_i)

    tpr, fpr = np.array(tpr), np.array(fpr)

    return fpr, tpr, thresholds


def roc_auc(**kwargs) -> float:
    """
    compute AUC (area under the curve):
    :param kwargs:
        set 1:
        fpr: numpy.ndarray(shape=(m), ), false positive rate;
        tpr: numpy.ndarray(shape=(m), ), true positive rate;
        set 2:
        true: numpy.ndarray(shape=(m), ), an array of true classes;
        predicted_prob: numpy.ndarray(shape=(m), ),
            an array of predicted probabilities of being the positive class;
        pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
    :return: auc, float, area under the curve;
    """
    if ('fpr' in kwargs) and ('tpr' in kwargs):
        fpr = kwargs['fpr']
        tpr = kwargs['tpr']
    elif ('true' in kwargs) and ('predicted_prob' in kwargs):
        # if fpr and tpr are not given: re-compute roc:
        if 'pos_label' in kwargs:
            pos_label = kwargs['pos_label']
        else:
            pos_label = _DEFAULT_POS_LABEL

        fpr, tpr, thresholds = roc(true=kwargs['true'], predicted_prob=kwargs['predicted_prob'], pos_label=pos_label)
    else:
        raise KeyError(f"given arguments are not sufficient to compute AUC.")

    # fpr, tpr = np.insert(fpr, 0, 1), np.insert(tpr, 0, 1)  # add 1 to the first position of the array;
    # fpr, tpr = np.append(fpr, 0), np.append(tpr, 0)  # add 0 to the last position of the arrays;

    fpr, tpr = np.insert(fpr, 0, 0), np.insert(tpr, 0, 0)  # add 0 to the first position of the array;
    fpr, tpr = np.append(fpr, 1), np.append(tpr, 1)  # add 1 to the last position of the arrays;
    # auc = float(np.sum(1 / 2 * (tpr[0:-1] + tpr[1:]) * (- np.diff(fpr))))  # reverse as roc is cal differently;
    auc = float(np.sum(1 / 2 * (tpr[0:-1] + tpr[1:]) * np.diff(fpr)))
    return auc


def roc_margins(**kwargs) -> dict:
    """
    compute differences between a fpr and tpr:
        also computes
    :param kwargs:
        set 1:
        fpr: numpy.ndarray(shape=(m), ), false positive rate;
        tpr: numpy.ndarray(shape=(m), ), true positive rate;
        set 2:
        true: numpy.ndarray(shape=(m), ), an array of true classes;
        predicted_prob: numpy.ndarray(shape=(m), ),
            an array of predicted probabilities of being the positive class;
        pos_label: [str, bool, int], positive class label, label that is considered as the positive class;
    :return: roc margin dict, dict, including:
        fpr: false positive rate, numpy.ndarray(shape=(m), );
        tpr, true positive rate, numpy.ndarray(shape=(m), );
        margins: margin curve, numpy.ndarray(shape=(m), );
        max_margin: the maximum margin between fpr and tpr;
        max_margin_threshold: threshold for such maximum margin;
    """

    if ('fpr' in kwargs) and ('tpr' in kwargs) and ('thresholds' in kwargs):
        fpr = kwargs['fpr']
        tpr = kwargs['tpr']
        thresholds = kwargs['thresholds']
    elif ('true' in kwargs) and ('predicted_prob' in kwargs):
        # if fpr and tpr are not given: re-compute roc:
        if 'pos_label' in kwargs:
            pos_label = kwargs['pos_label']
        else:
            pos_label = _DEFAULT_POS_LABEL

        fpr, tpr, thresholds = roc(true=kwargs['true'], predicted_prob=kwargs['predicted_prob'], pos_label=pos_label)
    else:
        raise KeyError(f"given arguments are not sufficient to compute AUC.")

    margins = tpr - fpr  # the margin between tpr and fpr
    max_margin = margins.max()
    max_margin_threshold = thresholds[margins.argmax()]

    roc_margin_dict = {'fpr': fpr,
                       'tpr': tpr,
                       'thresholds': thresholds,
                       'margins': margins,
                       'max_margin': max_margin,
                       'max_margin_threshold': max_margin_threshold}

    return roc_margin_dict



from typing import Union

import numpy as np
import pandas as pd


def data_type_converter(series: Union[pd.Series, np.ndarray, list]) -> np.ndarray:
    """
    convert list or pandas.Series to numpy.ndarray;
    :param series: [list, pandas.Series], a series;
    :return: numpy.ndarray(shape=(m), ), the converted array;
    """
    if isinstance(series, list):
        series = np.array(series)
    elif isinstance(series, np.ndarray):
        pass
    elif isinstance(series, pd.Series):
        series = series.values

    return series



# -*- coding: utf-8 -*-
# Author: XuMing <xuming624@qq.com>
# Brief:

import numpy as np
import sklearn.svm as svm
from sklearn.calibration import CalibratedClassifierCV
from sklearn.ensemble import BaggingClassifier, RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import PassiveAggressiveClassifier
from sklearn.linear_model import Perceptron
from sklearn.linear_model import SGDClassifier
from sklearn.multiclass import OneVsRestClassifier
from sklearn.naive_bayes import BernoulliNB
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn_crfsuite import CRF
from xgboost import XGBClassifier

n_estimators = 10


def get_model(model_type):
    return {
        'logreg': LogisticRegression(random_state=0, multi_class='auto'),
        'logreg_l2': LogisticRegression(penalty='l2', tol=0.0001, C=1.0),
        'logreg_l1': LogisticRegression(penalty='l1', tol=0.0001, C=1.0),

        'svm': OneVsRestClassifier(
            BaggingClassifier(svm.SVC(kernel="linear", probability=True), max_samples=1.0 / n_estimators,
                              n_estimators=n_estimators)),
        'knn': KNeighborsClassifier(
            n_neighbors=10,
            algorithm='brute',
            metric='cosine',
            n_jobs=2),
        'nb-multinomial': MultinomialNB(),
        'mlp': MLPClassifier(),
        'rf': RandomForestClassifier(max_depth=20, random_state=50, n_jobs=-1),
        'dt': DecisionTreeClassifier(
            random_state=789654,
            criterion="gini"),
        'xgb': XGBClassifier(
            max_depth=6,
            subsample=0.8,
            colsample_bytree=0.7,
            objective='multi:softmax',
            silent=True,
            booster='gbtree',
            learning_rate=0.05,
            n_jobs=-1),
        'linearsvm': CalibratedClassifierCV(svm.LinearSVC(dual=False)),
        'crf': CRF(
            algorithm='lbfgs',
            c1=1,
            c2=0.1,
            max_iterations=100,
            # include transitions that are possible, but not observed
            all_possible_transitions=True
        )
    }.get(model_type.lower(), None)


def get_online_model(model_type):
    return {
        'sgdclassifier': SGDClassifier(max_iter=5, loss='modified_huber'),
        'perceptron': Perceptron(),
        'nb-multinomial': MultinomialNB(alpha=0.01),
        'nb-bernoulli': BernoulliNB(alpha=0.01),
        'passive-aggressive': PassiveAggressiveClassifier(),
    }.get(model_type.lower(), None)


def get_model_with_parms(model_type, params):
    if model_type == "logistic_regression":
        model = LogisticRegression(params)
    elif model_type == "random_forest":
        model = RandomForestClassifier(params)
    elif model_type == "decision_tree":
        model = DecisionTreeClassifier(params)
    elif model_type == "knn":
        model = KNeighborsClassifier(params)
    elif model_type == "bayes":
        model = MultinomialNB(params)
    elif model_type == "xgboost":
        model = XGBClassifier(params)
    elif model_type == "svm":
        model = svm.SVC(params)
    elif model_type == 'mlp':
        model = MLPClassifier(params)

    return model


def get_model_parameters_range(model_type):
    paramgrid = {}
    if model_type == 'logistic_regression':
        paramgrid = {"penalty": ['l1', 'l2'],
                     "C": np.logspace(0, 4, 10)}
    elif model_type == 'svm':
        paramgrid = {"kernel": ["rbf", "linear"],
                     "C": np.logspace(-9, 9, num=25, base=10),
                     "gamma": np.logspace(-9, 9, num=25, base=10),
                     "probability": [True]}
    elif model_type == "decision_tree":
        paramgrid = {"criterion": ['gini', 'entropy'],
                     "min_samples": [2, 3, 4],
                     "class_weight": ['balanced', 'None']}
    elif model_type == "random_forest":
        paramgrid = {"n_estimators": [100, 200, 300, 400],
                     "max_features": ['auto', 'sqrt'],
                     "max_depth": np.linspace(10, 110, num=11),
                     "bootstrap": [True, False]
                     }
    elif model_type == "knn":
        paramgrid = {"n_neighbors": [4, 5, 6, 7, 8, 9, 10],
                     "algorithm": ['auto', 'ball_tree', 'kd_tree', 'brute'],
                     "p": [1, 2],
                     "weights": ['uniform', 'distance']}
    elif model_type == 'bayes':
        paramgrid = {"alpha": np.logspace(0, 2, 10),
                     "fit_prior": [True, False]}

    elif model_type == "xgboost":
        paramgrid = {'booster': ['gbtree', 'gblinear'], 'learning_rate': np.logspace(0.01, 0.5, 10),
                     'max_depth': [3, 4, 5, 6, 7, 8, 9, 10], 'gamma': np.logspace(0, 2, 10)}
    elif model_type == 'mlp':
        paramgrid = {'solver': ['lbfgs'], 'max_iter': [500, 1000, 1500], 'alpha': 10.0 ** -np.arange(1, 7),
                     'hidden_layer_sizes': np.arange(5, 12)}

    return paramgrid



# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# version: 1.0
# license: Apache Licence
# file: corpus.py
# time: 2019-05-17 11:28

import collections
import logging
import operator
from typing import List, Dict, Optional

import numpy as np
from tensorflow.keras.utils import to_categorical
from tensorflow.python.keras.preprocessing.sequence import pad_sequences

import kolibri.dnn
from kolibri.dnn import utils
from kolibri.indexers.base_indexer import BaseIndexer


class LabelingIndexer(BaseIndexer):
    """
    Corpus Pre Indexer class
    """

    def info(self):
        info = super(LabelingIndexer, self).info()
        info['task'] = kolibri.dnn.LABELING
        return info

    def build_label_dict(self,
                         label_list: List[List[str]]):
        """
        Build label2idx dict for sequence labeling task

        Args:
            label_list: corpus label list
        """
        label2idx: Dict[str: int] = {
            self.token_pad: 0
        }

        token2count = {}

        for sequence in label_list:
            for label in sequence:
                count = token2count.get(label, 0)
                token2count[label] = count + 1

        sorted_token2count = sorted(token2count.items(),
                                    key=operator.itemgetter(1),
                                    reverse=True)
        token2count = collections.OrderedDict(sorted_token2count)

        for token in token2count.keys():
            if token not in label2idx:
                label2idx[token] = len(label2idx)

        self.label2idx = label2idx
        self.idx2label = dict([(value, key)
                               for key, value in self.label2idx.items()])
        logging.debug(f"build label2idx dict finished, contains {len(self.label2idx)} labels.")

    def process_y_dataset(self,
                          data: List[List[str]],
                          max_len: Optional[int] = None,
                          subset: Optional[List[int]] = None) -> np.ndarray:
        if subset is not None:
            target = utils.get_list_subset(data, subset)
        else:
            target = data[:]
        numerized_samples = self.numerize_label_sequences(target)
        padded_seq = pad_sequences(
            numerized_samples, max_len, padding='post', truncating='post')
        return to_categorical(padded_seq, len(self.label2idx))

    def numerize_token_sequences(self,
                                 sequences: List[List[str]]):

        result = []
        for seq in sequences:
            if self.add_bos_eos:
                seq = [self.token_bos] + seq + [self.token_eos]
            unk_index = self.token2idx[self.token_unk]
            result.append([self.token2idx.get(token, unk_index) for token in seq])
        return result

    def numerize_label_sequences(self,
                                 sequences: List[List[str]]) -> List[List[int]]:
        result = []
        for seq in sequences:
            if self.add_bos_eos:
                seq = [self.token_pad] + seq + [self.token_pad]
            result.append([self.label2idx[label] for label in seq])
        return result

    def reverse_numerize_label_sequences(self,
                                         sequences,
                                         lengths=None):
        result = []

        for index, seq in enumerate(sequences):
            labels = []
            if self.add_bos_eos:
                seq = seq[1:]
            for idx in seq:
                labels.append(self.idx2label[idx])
            if lengths is not None:
                labels = labels[:lengths[index]]
            result.append(labels)
        return result


if __name__ == "__main__":
    from kolibri.data.corpus.corpus import CONLL2003ENCorpus

    x, y = CONLL2003ENCorpus.load_data()
    p = LabelingIndexer()
    p.analyze_corpus(x, y)
    r = p.process_x_dataset(x, subset=[10, 12, 20])
    print(r)



import logging
import os
import time
import warnings
from timeit import default_timer as timer

from numpy import unique
from sklearn.metrics import classification_report
from sklearn.model_selection import train_test_split

from kolibri.classifier_evaluator.metrics import confusion_matrix
from kolibri.config import TaskType
from kolibri.evaluators.base_evaluator import BaseEvaluator
from kolibri.utils import constants


class ModelEvaluator(BaseEvaluator):
    """ The prequential evaluation method, or interleaved test-then-train method,
    is an alternative to the traditional holdout evaluation, inherited from
    batch setting problems.
    """

    def __init__(self,
                 n_wait=200,
                 max_samples=100000,
                 batch_size=1,
                 pretrain_size=200,
                 train_ratio=0.8,
                 max_time=float("inf"),
                 metrics=None,
                 output_file=None,
                 show_plot=False,
                 is_stream=True,
                 data_points_for_classification=False):

        super().__init__()
        self._method = 'prequential'
        self.n_wait = n_wait
        self.max_samples = max_samples
        self.pretrain_size = pretrain_size
        self.batch_size = batch_size
        self.max_time = max_time
        self.output_file = output_file
        self.show_plot = show_plot
        self.train_ratio = train_ratio
        self.data_points_for_classification = data_points_for_classification

        if metrics is None and data_points_for_classification is False:
            self.metrics = [constants.ACCURACY, constants.KAPPA]

        elif data_points_for_classification is True:
            self.metrics = [constants.DATA_POINTS]

        else:
            self.metrics = metrics

        self.is_stream = is_stream
        self.n_sliding = n_wait

        warnings.filterwarnings("ignore", ".*invalid value encountered in true_divide.*")
        warnings.filterwarnings("ignore", ".*Passing 1d.*")

    def evaluate(self, stream, model, model_names=None, random_state=None):
        """ Evaluates a learner or set of learners on samples from a stream.

        Parameters
        ----------
        stream: Stream
            The stream from which to draw the samples.

        model: StreamModel or list
            The learner or list of learners to evaluate.

        model_names: list, optional (Default=None)
            A list with the names of the learners.

        Returns
        -------
        StreamModel or list
            The trained learner(s).

        """
        self._init_evaluation(model=model, stream=stream, model_names=model_names)

        if self._check_configuration():
            self._reset_globals()
            # Initialize metrics and outputs (plots, log files, ...)
            self._init_metrics()
            #            self._init_plot()
            self._init_file()
            if self.is_stream:
                self.model = self._train_and_test_stream()
            else:
                self.model = self._train_and_test(random_state)

            if self.show_plot:
                self.visualizer.hold()

            return self.model

    def _train_and_test_stream(self):
        """ Method to control the prequential evaluation.

        Returns
        -------
        BaseClassifier extension or list of BaseClassifier extensions
            The trained classification.

        Notes
        -----
        The classifier parameter should be an extension from the BaseClassifier. In
        the future, when BaseRegressor is created, it could be an extension from that
        class as well.

        """
        logging.basicConfig(format='%(message)s', level=logging.DEBUG)
        start_time = timer()
        end_time = timer()
        logging.info('Model Evaluation')
        logging.info('Evaluating %s label(s).', str(self.stream.n_targets))

        n_samples = self.stream.n_remaining_samples()
        if n_samples == -1 or n_samples > self.max_samples:
            n_samples = self.max_samples

        first_run = True
        if self.pretrain_size > 0:
            logging.info('Pre-training on %s samples.', str(self.pretrain_size))

            X, y = self.stream.next(self.pretrain_size)

            for i in range(self.n_models):
                self.model[i].partial_fit(X=X, y=y, classes=unique(self.stream.target_values))

            self.global_sample_count += self.pretrain_size
            first_run = False
        else:
            logging.info('No pre-training.')

        update_count = 0
        logging.info('Evaluating...')
        while ((self.global_sample_count < self.max_samples) & (end_time - start_time < self.max_time) & (
                self.stream.has_more_samples())):
            try:
                X, y = self.stream.next(self.batch_size)

                if X is not None and y is not None:
                    # Test
                    prediction = [[] for _ in range(self.n_models)]
                    for i in range(self.n_models):
                        try:
                            start = time.time()
                            prediction[i].extend(self.model[i].predict(X, return_labels=False))
                            print(time.time() - start)
                        except TypeError:
                            raise TypeError("Unexpected prediction value from {}"
                                            .format(type(self.model[i]).__name__))
                    self.global_sample_count += self.batch_size

                    for j in range(self.n_models):
                        for i in range(len(prediction[0])):
                            y_i = y[i]
                            if self._task_type != TaskType.MULTI_TARGET_REGRESSION:
                                y_i = self.model[j].numerize_label_sequences(y[i])
                            self.mean_eval_measurements[j].add_result(y_i, prediction[j][i])
                            self.current_eval_measurements[j].add_result(y_i, prediction[j][i])
                    self._check_progress(logging, n_samples)

                    # Train
                    if first_run:
                        for i in range(self.n_models):
                            if self._task_type != TaskType.REGRESSION and \
                                    self._task_type != TaskType.MULTI_TARGET_REGRESSION:
                                self.model[i].partial_fit(X, y, self.stream.target_values)
                            else:
                                self.model[i].partial_fit(X, y)
                        first_run = False
                    else:
                        for i in range(self.n_models):
                            self.model[i].partial_fit(X, y)

                    if ((self.global_sample_count % self.n_wait) == 0 or
                            (self.global_sample_count >= self.max_samples) or
                            (self.global_sample_count / self.n_wait > update_count + 1)):
                        if prediction is not None:
                            self._update_metrics()
                        update_count += 1

                end_time = timer()
            except BaseException as exc:
                print(exc)
                if exc is KeyboardInterrupt:
                    self._update_metrics()
                break

        self.evaluation_summary(logging, start_time, end_time)

        if self.is_stream:
            self.stream.restart()

        return self.model

    def _train_and_test(self, random_state=None):
        """ Method to control the prequential evaluation.

        Returns
        -------
        BaseClassifier extension or list of BaseClassifier extensions
            The trained classification.

        Notes
        -----
        The classifier parameter should be an extension from the BaseClassifier. In
        the future, when BaseRegressor is created, it could be an extension from that
        class as well.

        """
        logging.basicConfig(format='%(message)s', level=logging.DEBUG)
        start_time = timer()
        end_time = timer()
        logging.info('Model Evaluation')
        logging.info('Evaluating %s label(s).', str(self.stream.n_targets))

        n_samples = self.stream.n_remaining_samples()
        if n_samples == -1 or n_samples > self.max_samples:
            n_samples = self.max_samples

        X, y = self.stream.next(n_samples)

        X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=self.train_ratio,
                                                            random_state=random_state)
        for i in range(self.n_models):
            self.model[i].fit(X=X_train, y=y_train)

        update_count = 0
        logging.info('Evaluating...')
        prediction = [[] for _ in range(self.n_models)]
        for i in range(self.n_models):
            try:
                prediction[i] = self.model[i].predict(X_test)
            except TypeError:
                raise TypeError("Unexpected prediction value from {}".format(type(self.model[i]).__name__))

        for j in range(self.n_models):
            print(classification_report(y_test, prediction[j]))
            print(confusion_matrix(y_test, prediction[j]))
        #           confusion_matrix_plot.plot_confusion_matrix_by_dict(confusion_matrix(y_test, prediction[j]))

        if self.is_stream:
            self.stream.restart()

        return self.model

    def _periodic_holdout(self):
        """ Method to control the holdout evaluation.
    The holdout evaluation method or periodic holdout evaluation method.

    Analyses each arriving sample by updating its statistics, without computing
    performance metrics, nor predicting y_values or regression values.

    The performance evaluation happens at every n_wait analysed samples, at which
    moment the evaluator will test the learners performance on a test set, formed
    by yet unseen samples, which will be used to evaluate performance, but not to
    train the model.

    It's possible to use the same test set for every test made or to dynamically
    create test sets, so that they differ from each other. If dynamic test sets
    are enabled, we use the texts stream to create test sets on the go. This process
    is more likely to generate test sets that follow the current concept, in
    comparison to static test sets.

    Thus, if concept drift is known to be present in the stream, using dynamic
    test sets is recommended. If no concept drift is expected, disabling this
    parameter will speed up the evaluation process.
        """
        self._start_time = timer()
        self._end_time = timer()
        print('Holdout Evaluation')
        print('Evaluating {} label(s).'.format(self.stream.n_targets))

        actual_max_samples = self.stream.n_remaining_samples()
        if actual_max_samples == -1 or actual_max_samples > self.max_samples:
            actual_max_samples = self.max_samples

        first_run = True

        if not self.dynamic_test_set:
            print('Separating {} holdout samples.'.format(self.test_size))
            self.X_test, self.y_test = self.stream.next(self.test_size)
            self.global_sample_count += self.test_size

        performance_sampling_cnt = 0
        print('Evaluating...')
        while ((self.global_sample_count < self.max_samples) & (self._end_time - self._start_time < self.max_time)
               & (self.stream.has_more_samples())):
            try:
                X, y = self.stream.next(self.batch_size)

                if X is not None and y is not None:
                    self.global_sample_count += self.batch_size

                    # Train
                    if first_run:
                        for i in range(self.n_models):
                            if self._task_type == constants.CLASSIFICATION:
                                self.current_eval_measurements[i].compute_training_time_begin()
                                self.model[i].partial_fit(X=X, y=y, classes=self.stream.target_values)
                                self.current_eval_measurements[i].compute_training_time_end()
                            elif self._task_type == constants.MULTI_TARGET_CLASSIFICATION:
                                self.current_eval_measurements[i].compute_training_time_begin()
                                self.model[i].partial_fit(X=X, y=y, classes=unique(self.stream.target_values))
                                self.current_eval_measurements[i].compute_training_time_end()
                            else:
                                self.current_eval_measurements[i].compute_training_time_begin()
                                self.model[i].partial_fit(X=X, y=y)
                                self.current_eval_measurements[i].compute_training_time_end()
                            self.current_eval_measurements[i].update_time_measurements(self.batch_size)
                        first_run = False
                    else:
                        for i in range(self.n_models):
                            # Compute running time
                            self.current_eval_measurements[i].compute_training_time_begin()
                            self.model[i].partial_fit(X, y)
                            self.current_eval_measurements[i].compute_training_time_end()
                            self.current_eval_measurements[i].update_time_measurements(self.batch_size)

                    self._check_progress(actual_max_samples)  # TODO Confirm place

                    # Test on holdout set
                    if self.dynamic_test_set:
                        perform_test = self.global_sample_count >= (self.n_wait * (performance_sampling_cnt + 1)
                                                                    + (self.test_size * performance_sampling_cnt))
                    else:
                        perform_test = (self.global_sample_count - self.test_size) % self.n_wait == 0

                    if perform_test | (self.global_sample_count >= self.max_samples):

                        if self.dynamic_test_set:
                            print('Separating {} holdout samples.'.format(self.test_size))
                            self.X_test, self.y_test = self.stream.next(self.test_size)
                            self.global_sample_count += get_dimensions(self.X_test)[0]

                        # Test
                        if (self.X_test is not None) and (self.y_test is not None):
                            prediction = [[] for _ in range(self.n_models)]
                            for i in range(self.n_models):
                                try:
                                    self.current_eval_measurements[i].compute_testing_time_begin()
                                    prediction[i].extend(self.model[i].predict(self.X_test, return_labels=False))
                                    self.current_eval_measurements[i].compute_testing_time_end()
                                    self.current_eval_measurements[i].update_time_measurements(self.test_size)
                                except TypeError:
                                    raise TypeError("Unexpected prediction value from {}"
                                                    .format(type(self.model[i]).__name__))
                            if prediction is not None:
                                for j in range(self.n_models):
                                    for i in range(len(prediction[0])):
                                        self.mean_eval_measurements[j].add_result(self.y_test[i],
                                                                                  prediction[j][i])
                                        self.current_eval_measurements[j].add_result(self.y_test[i],
                                                                                     prediction[j][i])

                                self._update_metrics()
                            performance_sampling_cnt += 1

                self._end_time = timer()
            except BaseException as exc:
                print(exc)
                if exc is KeyboardInterrupt:
                    self._update_metrics()
                break

        # Flush file buffer, in case it contains texts
        self._flush_file_buffer()

        self.evaluation_summary()

        if self.restart_stream:
            self.stream.restart()

        return self.model

    def set_params(self, parameter_dict):
        """ This function allows the users to change some of the evaluator's parameters,
        by passing a dictionary where keys are the parameters names, and values are
        the new parameters' values.

        Parameters
        ----------
        parameter_dict: Dictionary
            A dictionary where the keys are the names of attributes the user
            wants to change, and the values are the new values of those attributes.

        """
        for name, value in parameter_dict.items():
            if name == 'n_wait':
                self.n_wait = value
            elif name == 'max_samples':
                self.max_samples = value
            elif name == 'pretrain_size':
                self.pretrain_size = value
            elif name == 'batch_size':
                self.batch_size = value
            elif name == 'max_time':
                self.max_time = value
            elif name == 'output_file':
                self.output_file = value
            elif name == 'show_plot':
                self.show_plot = value

    def get_info(self):
        filename = "None"
        if self.output_file is not None:
            _, filename = os.path.split(self.output_file)
        return 'Prequential Evaluator: n_wait: ' + str(self.n_wait) + \
               ' - max_samples: ' + str(self.max_samples) + \
               ' - max_time: ' + str(self.max_time) + \
               ' - output_file: ' + filename + \
               ' - batch_size: ' + str(self.batch_size) + \
               ' - pretrain_size: ' + str(self.pretrain_size) + \
               ' - task_type: ' + self._task_type + \
               ' - show_plot: ' + ('True' if self.show_plot else 'False') + \
               ' - metrics: ' + (str(self.metrics) if self.metrics is not None else 'None')



import logging

from kolibri.application.datastore.storage_adapter import DatabaseAdapter

logger = logging.getLogger(__name__)


class SQLDatabaseAdapter(DatabaseAdapter):
    """
    The SQLDatabaseAdapter allows kolibri to store documents
    texts in any database supported by the SQL Alchemy ORM.

    All parameters are optional, by default a sqlite database is used.

    :keyword database_uri: eg: sqlite:///database_test.sqlite3',
        The database_uri can be specified to choose database driver.
    :type database_uri: str
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        from sqlalchemy import create_engine
        from sqlalchemy.orm import sessionmaker

        self.database_uri = kwargs.get('database_uri', False)

        # None results in a sqlite in-memory database as the default
        if self.database_uri is None:
            self.database_uri = 'sqlite://'

        # Create a file database if the database is not a connection string
        if not self.database_uri:
            self.database_uri = 'sqlite:///db.sqlite3'

        self.engine = create_engine(self.database_uri, convert_unicode=True)

        if self.database_uri.startswith('sqlite://'):
            from sqlalchemy.engine import Engine
            from sqlalchemy import event

            @event.listens_for(Engine, 'connect')
            def set_sqlite_pragma(dbapi_connection, connection_record):
                dbapi_connection.execute('PRAGMA journal_mode=WAL')
                dbapi_connection.execute('PRAGMA synchronous=NORMAL')

        if not self.engine.dialect.has_table(self.engine, 'Document'):
            self.create_database()

        self.Session = sessionmaker(bind=self.engine, expire_on_commit=True)

    def get_document_model(self):
        """
        Return the document model.
        """
        from kolibri.application.datastore.sqlalchemy_models.models import Document
        return Document

    def get_tag_model(self):
        """
        Return the document model.
        """
        from kolibri.application.datastore.sqlalchemy_models.models import Tag
        return Tag

    def model_to_object(self, document):
        from kolibri.document import Document as DocumentObject

        return DocumentObject(**document.serialize())

    def count(self):
        """
        Return the number of entries in the database.
        """
        Document = self.get_model('document')

        session = self.Session()
        document_count = session.query(Document).count()
        session.close()
        return document_count

    def remove(self, document_text):
        """
        Removes the document that matches the input text.
        Removes any responses from documents where the response text matches
        the input text.
        """
        Document = self.get_model('document')
        session = self.Session()

        query = session.query(Document).filter_by(text=document_text)
        record = query.first()

        session.delete(record)

        self._session_finish(session)

    def filter(self, **kwargs):
        """
        Returns a list of objects from the database.
        The kwargs parameter can contain any number
        of attributes. Only objects which contain all
        listed attributes and in which all values match
        for all listed attributes will be returned.
        """
        from sqlalchemy import or_

        Document = self.get_model('document')
        Tag = self.get_model('tag')

        session = self.Session()

        page_size = kwargs.pop('page_size', 1000)
        order_by = kwargs.pop('order_by', None)
        tags = kwargs.pop('tags', [])
        exclude_text = kwargs.pop('exclude_text', None)
        exclude_text_words = kwargs.pop('exclude_text_words', [])
        search_text_contains = kwargs.pop('search_text_contains', None)

        # Convert a single sting into a list if only one tag is provided
        if type(tags) == str:
            tags = [tags]

        if len(kwargs) == 0:
            documents = session.query(Document).filter()
        else:
            documents = session.query(Document).filter_by(**kwargs)

        if tags:
            documents = documents.join(Document.tags).filter(
                Tag.name.in_(tags)
            )

        if exclude_text:
            documents = documents.filter(
                ~Document.text.in_(exclude_text)
            )

        if exclude_text_words:
            or_word_query = [
                Document.text.ilike('%' + word + '%') for word in exclude_text_words
            ]
            documents = documents.filter(
                ~or_(*or_word_query)
            )

        if search_text_contains:
            or_query = [
                Document.search_text.contains(word) for word in search_text_contains.split(' ')
            ]
            documents = documents.filter(
                or_(*or_query)
            )

        if order_by:

            if 'created_at' in order_by:
                index = order_by.index('created_at')
                order_by[index] = Document.created_at.asc()

            documents = documents.order_by(*order_by)

        total_documents = documents.count()

        for start_index in range(0, total_documents, page_size):
            for document in documents.slice(start_index, start_index + page_size):
                yield self.model_to_object(document)

        session.close()

    def create(self, **kwargs):
        """
        Creates a new document matching the keyword arguments specified.
        Returns the created document.
        """
        Document = self.get_model('document')
        Tag = self.get_model('tag')

        session = self.Session()

        tags = set(kwargs.pop('tags', []))

        if 'search_text' not in kwargs:
            kwargs['search_text'] = self.tagger.get_text_index_string(kwargs['text'])

        document = Document(**kwargs)

        for tag_name in tags:
            tag = session.query(Tag).filter_by(name=tag_name).first()

            if not tag:
                # Create the tag
                tag = Tag(name=tag_name)

            document.tags.append(tag)

        session.add(document)

        session.flush()

        session.refresh(document)

        document_object = self.model_to_object(document)

        self._session_finish(session)

        return document_object

    def create_many(self, documents):
        """
        Creates multiple document entries.
        """
        Document = self.get_model('document')
        Tag = self.get_model('tag')

        session = self.Session()

        create_documents = []
        create_tags = {}

        for document in documents:

            document_data = document.serialize()
            tag_data = document_data.pop('tags', [])

            document_model_object = Document(**document_data)

            if not document.search_text:
                document_model_object.search_text = self.tagger.get_text_index_string(document.text)

            if not document.search_target_text and document.target_text:
                document_model_object.search_target_text = self.tagger.get_text_index_string(document.target_text)

            new_tags = set(tag_data) - set(create_tags.keys())

            if new_tags:
                existing_tags = session.query(Tag).filter(
                    Tag.name.in_(new_tags)
                )

                for existing_tag in existing_tags:
                    create_tags[existing_tag.name] = existing_tag

            for tag_name in tag_data:
                if tag_name in create_tags:
                    tag = create_tags[tag_name]
                else:
                    # Create the tag if it does not exist
                    tag = Tag(name=tag_name)

                    create_tags[tag_name] = tag

                document_model_object.tags.append(tag)
            create_documents.append(document_model_object)

        session.add_all(create_documents)
        session.commit()

    def update(self, document):
        """
        Modifies an entry in the database.
        Creates an entry if one does not exist.
        """
        Document = self.get_model('document')
        Tag = self.get_model('tag')

        if document is not None:
            session = self.Session()
            record = None

            if hasattr(document, 'id') and document.id is not None:
                record = session.query(Document).get(document.id)
            else:
                record = session.query(Document).filter(
                    Document.text == document.text,
                    Document.document == document.document,
                ).first()

                # Create a new document entry if one does not already exist
                if not record:
                    record = Document(
                        text=document.text,
                        document=document.document,
                        persona=document.persona
                    )

            # Update the response value
            record.in_response_to = document.in_response_to

            record.created_at = document.created_at

            record.search_text = self.tagger.get_text_index_string(document.text)

            if document.in_response_to:
                record.search_target_text = self.tagger.get_text_index_string(document.in_response_to)

            for tag_name in document.get_tags():
                tag = session.query(Tag).filter_by(name=tag_name).first()

                if not tag:
                    # Create the record
                    tag = Tag(name=tag_name)

                record.tags.append(tag)

            session.add(record)

            self._session_finish(session)

    def get_random(self):
        """
        Returns a random document from the database.
        """
        import random

        Document = self.get_model('document')

        session = self.Session()
        count = self.count()
        if count < 1:
            raise Exception('database is empty')

        random_index = random.randrange(0, count)
        random_document = session.query(Document)[random_index]

        document = self.model_to_object(random_document)

        session.close()
        return document

    def drop(self):
        """
        Drop the database.
        """
        Document = self.get_model('document')
        Tag = self.get_model('tag')

        session = self.Session()

        session.query(Document).delete()
        session.query(Tag).delete()

        session.commit()
        session.close()

    def create_database(self):
        """
        Populate the database with the tables.
        """
        from kolibri.application.datastore.sqlalchemy_models.models import Base
        Base.metadata.create_all(self.engine)

    def _session_finish(self, session, document_text=None):
        from sqlalchemy.exc import InvalidRequestError
        try:
            session.commit()
        except InvalidRequestError:
            # Log the document text and the exception
            logger.exception(document_text)
        finally:
            session.close()



from tensorflow.keras import backend as K
from tensorflow.keras.layers import Lambda, Input, Dense
from tensorflow.keras.losses import mse, binary_crossentropy
from tensorflow.keras.models import Model

from kolibri.dnn.autoencoder.base_autoencoder import BaseAutoencoder


class VariationalAutoencoder(BaseAutoencoder):

    def sampling(self, args):
        """Reparameterization trick by sampling
            fr an isotropic unit Gaussian.

        # Arguments:
            args (tensor): mean and log of variance of Q(z|X)

        # Returns:
            z (tensor): sampled latent vector
        """

        z_mean, z_log_var = args
        # K is the keras backend
        batch = K.shape(z_mean)[0]
        dim = K.int_shape(z_mean)[1]
        # by default, random_normal has mean=0 and std=1.0
        epsilon = K.random_normal(shape=(batch, dim))
        return z_mean + K.exp(0.5 * z_log_var) * epsilon

    @classmethod
    def get_default_hyper_parameters(cls):
        return {
            'layer_multiplyer': 2,
            'input_distribution': 'gaussian',
            'hidden': {
                'units': 100,
                'activation': 'relu'
            },
            'latent':{
                'units': 10,
                'activation': 'relu'
                       },
            'output': {
                'activation': 'tanh'
                }
                }

    def build_model_arc(self):
        # network parameters
        input_shape = (self.input_dim,)
        latent_dim = self.hyper_parameters['latent']['units']

        # VAE model = encoder + decoder
        # build encoder model
        inputs = Input(shape=input_shape, name='encoder_input')
        x = Dense(**self.hyper_parameters['hidden'])(inputs)
        z_mean = Dense(**self.hyper_parameters['latent'], name='z_mean')(x)
        z_log_var = Dense(**self.hyper_parameters['latent'], name='z_log_var')(x)

        # use reparameterization trick to push the sampling out as input
        # note that "output_shape" isn't necessary
        # with the TensorFlow backend
        z = Lambda(self.sampling,
                   output_shape=(latent_dim,),
                   name='z')([z_mean, z_log_var])

        # instantiate encoder model
        self.encoder = Model(inputs, [z_mean, z_log_var, z], name='encoder')


        # decoder model
        latent_inputs = Input(shape=(latent_dim,), name='z_sampling')
        x = Dense(**self.hyper_parameters['hidden'])(latent_inputs)
        outputs = Dense(self.input_dim, activation='sigmoid')(x)

        # instantiate decoder model
        self.decoder = Model(latent_inputs, outputs, name='decoder')

        # instantiate VAE model
        outputs = self.decoder(self.encoder(inputs)[2])
        self.autoencoder = Model(inputs, outputs, name='vae_mlp')

        # VAE loss = mse_loss or xent_loss + kl_loss
        if self.hyper_parameters['input_distribution']=='bernoulli':
            reconstruction_loss = binary_crossentropy(inputs,
                                                      outputs)
        else:
            reconstruction_loss = mse(inputs, outputs)

        reconstruction_loss *= self.input_dim
        kl_loss = 1 + z_log_var - K.square(z_mean) - K.exp(z_log_var)
        kl_loss = K.sum(kl_loss, axis=-1)
        kl_loss *= -0.5
        vae_loss = K.mean(reconstruction_loss + kl_loss)
        self.autoencoder.add_loss(vae_loss)

if __name__ == "__main__":


    from kolibri.data.corpus.corpus import CreditCardFraud



    corpus = CreditCardFraud()
    train_x=corpus.X

    model = Autoencoder()
    model.build_model(train_x[:100])



    epochs = 4
    history = None
    model.fit(train_x[0:20000], epochs=epochs)

    model.evaluate(train_x[0:10000])

